# import typing
import unittest
import logging

from tkmilan.validation import StaticList, StaticMap, StaticMapLabels, StaticMapValues
from tkmilan.validation import LimitBounded, LimitUnbounded
from tkmilan import fn

logger = logging.getLogger(__name__)


class Test_StaticList(unittest.TestCase):
    def test_valid(self):
        spec_basic = StaticList(['1', '2', '3'], default='2')
        with self.subTest(spec=spec_basic):
            self.assertTrue('2' in spec_basic)
            self.assertFalse('10' in spec_basic)
            self.assertEqual(len(spec_basic), 3)

    def test_errors(self):
        with self.assertRaises(ValueError):
            StaticList(['1', '2', '3'], default='5')
        with self.assertRaises(ValueError):
            StaticList(['1', '2', '3'], defaultIndex=5)


class Test_StaticMap(unittest.TestCase):
    map_basic = {'1': 1, '2': 2, '3': 3}

    def test_valid(self):
        spec_basic = StaticMap(self.map_basic, defaultLabel='2')
        with self.subTest(spec=spec_basic):
            self.assertTrue('2' in spec_basic)
            self.assertTrue(2 in spec_basic.rlabels)
            self.assertFalse('5' in spec_basic)
            self.assertFalse(5 in spec_basic.rlabels)
            self.assertEqual(len(spec_basic), 3)

    def test_valid_wrappers(self):
        spec_basic_labels = StaticMapLabels(int, tuple(self.map_basic.keys()), defaultIndex=0)
        with self.subTest(spec=spec_basic_labels):
            self.assertTrue('2' in spec_basic_labels)
            self.assertTrue(2 in spec_basic_labels.rlabels)
            self.assertFalse('5' in spec_basic_labels)
            self.assertFalse(5 in spec_basic_labels.rlabels)
            self.assertTrue('1', spec_basic_labels.ldefault)
        spec_basic_values = StaticMapValues(str, tuple(self.map_basic.values()), defaultIndex=-1)
        with self.subTest(spec=spec_basic_values):
            self.assertTrue('2' in spec_basic_values)
            self.assertTrue(2 in spec_basic_values.rlabels)
            self.assertFalse('5' in spec_basic_values)
            self.assertFalse(5 in spec_basic_values.rlabels)
            self.assertTrue('3', spec_basic_values.ldefault)

    def test_errors(self):
        with self.assertRaises(ValueError):
            StaticMap(self.map_basic, defaultLabel='5')
        with self.assertRaises(ValueError):
            StaticMap(self.map_basic, defaultValue=5)
        with self.assertRaises(AssertionError):
            StaticMapLabels(str, tuple(self.map_basic.values()), defaultIndex=0)


class Test_Limit(unittest.TestCase):
    def test_int_bounded(self):
        lim_imin = LimitBounded(1, 10, fn=fn.valNumber, imin=True, imax=False)
        with self.subTest(limit=str(lim_imin)):
            self.assertTrue('1' in lim_imin)
            self.assertTrue('5' in lim_imin)
            self.assertFalse('10' in lim_imin)
        lim_imax = LimitBounded('1', '10', fn=fn.valNumber, imin=False, imax=True)
        with self.subTest(limit=str(lim_imax)):
            self.assertFalse('1' in lim_imax)
            self.assertTrue('5' in lim_imax)
            self.assertTrue('10' in lim_imax)
        lim_inone = LimitBounded('1', '10', fn=fn.valNumber, imin=False, imax=False, default='5')
        with self.subTest(limit=str(lim_inone)):
            self.assertFalse('1' in lim_inone)
            self.assertTrue('5' in lim_inone)
            self.assertFalse('10' in lim_inone)

    def test_int_bounded_step(self):
        lim_imin = LimitBounded(0, 100, step=10, fn=fn.valNumber,
                                imin=True, imax=False)
        with self.subTest(limit=lim_imin):
            self.assertTrue('0' in lim_imin)
            self.assertFalse('1' in lim_imin)
            self.assertFalse('49' in lim_imin)
            self.assertTrue('50' in lim_imin)
            self.assertFalse('51' in lim_imin)
            self.assertFalse('100' in lim_imin)
        lim_imax = LimitBounded('33', '39', step=3, fn=fn.valNumber,
                                imin=False, imax=True)
        with self.subTest(limit=lim_imax):
            self.assertFalse('33' in lim_imax)
            self.assertTrue('36' in lim_imax)
            self.assertTrue('39' in lim_imax)
        lim_inone = LimitBounded(19, 31, step=2, fn=fn.valNumber,
                                 imin=False, imax=False, default='25')
        with self.subTest(limit=lim_inone):
            self.assertFalse('19' in lim_inone)
            self.assertFalse('20' in lim_inone)
            self.assertTrue('21' in lim_inone)
            self.assertTrue('25' in lim_inone)
            self.assertTrue('29' in lim_inone)
            self.assertFalse('30' in lim_inone)
            self.assertFalse('31' in lim_inone)
            self.assertListEqual(list(lim_inone), ['21', '23', '25', '27', '29'])
        lim_iboth = LimitBounded(12, 28, step=2, fn=fn.valNumber,
                                 imin=True, imax=True)
        with self.subTest(limit=lim_iboth):
            self.assertTrue('12' in lim_iboth)
            self.assertTrue('28' in lim_iboth)
            self.assertEqual(len(lim_iboth), 1 + (28 - 12) // 2)

    def test_int_infinite(self):
        lim_nomin = LimitUnbounded(None, 10, fn=fn.valNumber)
        with self.subTest(limit=str(lim_nomin)):
            self.assertTrue('-100' in lim_nomin)
            self.assertTrue('5' in lim_nomin)
            self.assertFalse('+100' in lim_nomin)
        lim_nomax = LimitUnbounded('1', None, fn=fn.valNumber)
        with self.subTest(limit=str(lim_nomax)):
            self.assertFalse('-100' in lim_nomax)
            self.assertTrue('5' in lim_nomax)
            self.assertTrue('+100' in lim_nomax)

    def test_errors(self):
        for cls in (LimitBounded, LimitUnbounded):
            with self.assertRaises(ValueError):
                cls('1', '10', fn=fn.valNumber, imin=False, imax=False)  # Default Default 0 not in range
            with self.assertRaises(ValueError):
                # Weirdness: Strange `fn`
                cls('x', 'xxx', fn=len, default=2)  # No default roundtrip: str(2) == '2'; len('2') == 1; 1 != 2
        # Steps
        with self.assertRaises(ValueError):
            LimitBounded(1, 10, step=2, fn=fn.valNumber)  # max=6
        with self.assertRaises(ValueError):
            LimitBounded(10, 59, step=5, imax=False,      # max=60
                         fn=fn.valNumber)
        with self.assertRaises(ValueError):
            LimitUnbounded('-5', '5', step=1, fn=fn.valNumber)  # step:unsupported

    def test_padsize(self):
        lim_basic = LimitBounded(1, 0xFF, fn=fn.valNumber)
        with self.subTest(limit=str(lim_basic)):
            self.assertEqual(lim_basic.count_padsize(2), 8)
            self.assertEqual(lim_basic.count_padsize(16), 2)
        for n in range(30):
            for num in (2**n - n, 2**n, 2**n + n):
                lim = LimitBounded(1, num, fn=fn.valNumber)
                with self.subTest(limit=str(lim)):
                    self.assertEqual(lim.count_padsize(2), len(bin(num)) - 2)  # 0b
                    self.assertEqual(lim.count_padsize(8), len(oct(num)) - 2)  # 0o
                    self.assertEqual(lim.count_padsize(10), len(str(num)))
                    self.assertEqual(lim.count_padsize(16), len(hex(num)) - 2)  # 0x


if __name__ == '__main__':
    import sys
    logs_lvl = logging.DEBUG if '-v' in sys.argv else logging.INFO
    logging.basicConfig(level=logs_lvl, format='%(levelname)5.5s:%(funcName)s: %(message)s', stream=sys.stderr)
    unittest.main()



import typing
import unittest
import logging

from defaultlist import defaultlist

from tkmilan.model import Direction, VState, NotebookTabOrder

logger = logging.getLogger(__name__)


class Test_Direction_Grid(unittest.TestCase):
    GOLDEN = {
        ((1, 2), None): {
            Direction.N: ((0, 1),),
            Direction.S: ((0, 1),),
            Direction.V: ((1, 0),),
            Direction.E: ((0, 1),),
            Direction.W: ((1, 0),),
            Direction.H: ((0, 1),),
        },
        ((2, 1), None): {
            Direction.N: ((1,), (0,)),
            Direction.S: ((0,), (1,)),
            Direction.V: ((0,), (1,)),
            Direction.E: ((0,), (1,)),
            Direction.W: ((0,), (1,)),
            Direction.H: ((1,), (0,)),
        },
        ((2, 2), None): {
            Direction.N: ((1, 3), (0, 2)),
            Direction.S: ((0, 2), (1, 3)),
            Direction.V: ((2, 0), (3, 1)),
            Direction.E: ((0, 1), (2, 3)),
            Direction.W: ((1, 0), (3, 2)),
            Direction.H: ((2, 3), (0, 1)),
        },
        ((3, 3), None): {
            Direction.N: ((2, 5, 8), (1, 4, 7), (0, 3, 6)),
            Direction.S: ((0, 3, 6), (1, 4, 7), (2, 5, 8)),
            Direction.V: ((6, 3, 0), (7, 4, 1), (8, 5, 2)),
            Direction.E: ((0, 1, 2), (3, 4, 5), (6, 7, 8)),
            Direction.W: ((2, 1, 0), (5, 4, 3), (8, 7, 6)),
            Direction.H: ((6, 7, 8), (3, 4, 5), (0, 1, 2)),
        },
        # With amounts
        ((2, 2), 3): {
            Direction.N: ((1, 2), (0, 2)),
            Direction.S: ((0, 2), (1, 2)),
            Direction.V: ((2, 0), (2, 1)),
            Direction.E: ((0, 1), (2, 2)),
            Direction.W: ((1, 0), (2, 2)),
            Direction.H: ((2, 2), (0, 1)),
        },
        ((3, 3), 7): {
            Direction.N: ((2, 5, 6), (1, 4, 6), (0, 3, 6)),
            Direction.S: ((0, 3, 6), (1, 4, 6), (2, 5, 6)),
            Direction.V: ((6, 3, 0), (6, 4, 1), (6, 5, 2)),
            Direction.E: ((0, 1, 2), (3, 4, 5), (6, 6, 6)),
            Direction.W: ((2, 1, 0), (5, 4, 3), (6, 6, 6)),
            Direction.H: ((6, 6, 6), (3, 4, 5), (0, 1, 2)),
        },
    }
    ROTTEN = (  # Direction, *args, amount
        # Empty Columns
        (Direction.N, (2, 5), 7),
        (Direction.S, (2, 5), 7),
        # Empty Rows
        (Direction.E, (5, 2), 7),
        (Direction.W, (5, 2), 7),
    )

    def runTest(self):
        logger.info('Testing "%s"', self.__class__.__qualname__)
        for ((r, c), amount), golden in self.GOLDEN.items():
            for d in Direction:
                with self.subTest(r=r, c=c, amount=amount, d=d):
                    matrix: typing.List[typing.List[typing.Optional[int]]] = [[None for _ in range(c)] for _ in range(r)]
                    for idx, gc in enumerate(d.grid(r, c, amount=amount)):
                        for dr in range(gc.rowspan):
                            for dc in range(gc.columnspan):
                                matrix[gc.row + dr][gc.column + dc] = idx
                    logger.info('» %dx%d%s %s', r, c, '[%d]' % amount if amount else '', d.name)
                    for row in matrix:
                        logger.info('| %s', ' '.join(('x' * 2 if i is None else '%02d' % i for i in row)))
                    matrix_tuple = tuple(map(tuple, matrix))
                    if d not in golden:
                        self.fail('Missing @ %dx%d: `Direction.%s: %r,`' % (r, c, d.name, matrix_tuple))
                    self.assertEqual(golden[d], matrix_tuple, 'Error when calculation coordinates')
        for (d, (r, c), amount) in self.ROTTEN:
            with self.subTest(r=r, c=c, amount=amount, d=d):
                with self.assertRaises(ValueError) as e:
                    list(d.grid(r, c, amount=amount))
                logger.info('» %dx%d%s %s', r, c, '[%d]' % amount if amount else '', d.name)
                logger.info('  %s', e.exception)


class Test_Direction_Multiple(unittest.TestCase):
    GOLDEN = {
        ((1, 3), None): {  # `$1,3`
            Direction.N: ((0, 3), (0, 2), (0, 1)),
            Direction.S: ((0, 1), (0, 2), (0, 3)),
            Direction.V: ((1, 0), (2, 0), (3, 0)),
            Direction.E: ((0, 0, 0), (1, 2, 3)),
            Direction.W: ((0, 0, 0), (3, 2, 1)),
            Direction.H: ((1, 2, 3), (0, 0, 0)),
        },
        ((2, None, 2), 7): {  # `$2,x,2`
            Direction.N: ((1, 4, 6), (1, 4, 6), (1, 3, 6), (0, 3, 5), (0, 2, 5), (0, 2, 5)),
            Direction.S: ((0, 2, 5), (0, 2, 5), (0, 3, 5), (1, 3, 6), (1, 4, 6), (1, 4, 6)),
            Direction.V: ((5, 2, 0), (5, 2, 0), (5, 3, 0), (6, 3, 1), (6, 4, 1), (6, 4, 1)),
            Direction.E: ((0, 0, 0, 1, 1, 1), (2, 2, 3, 3, 4, 4), (5, 5, 5, 6, 6, 6)),
            Direction.W: ((1, 1, 1, 0, 0, 0), (4, 4, 3, 3, 2, 2), (6, 6, 6, 5, 5, 5)),
            Direction.H: ((5, 5, 5, 6, 6, 6), (2, 2, 3, 3, 4, 4), (0, 0, 0, 1, 1, 1)),
        },
    }
    ROTTEN = (  # Direction, amounts, amount
        # Non-Integer amount to distribute
        (None, (None, None), 3),
        (None, (None, 1, None), 4),  # TODO: This could be supported
        # Bad Distributions
        # - Not enough widgets to distribute: `$1,3,x`
        (None, (1, 3, None), 4),
    )

    def runTest(self):
        logger.info('Testing "%s"', self.__class__.__qualname__)
        for (amounts, amount), golden in self.GOLDEN.items():
            for d in Direction:
                with self.subTest(d=d, amounts=amounts, amount=amount):
                    matrix = defaultlist(lambda: defaultlist())
                    for idx, gc in enumerate(d.multiples(*amounts, amount=amount)):
                        for dr in range(gc.rowspan):
                            for dc in range(gc.columnspan):
                                matrix[gc.row + dr][gc.column + dc] = idx
                    logger.info('» %s (%s)%s', d.name, ' '.join((str(a) if a else 'x' for a in amounts)), '[%d]' % amount if amount else '')
                    for row in matrix:
                        logger.info('| %s', ' '.join(('x' * 2 if i is None else '%02d' % i for i in row)))
                    matrix_tuple = tuple(map(tuple, matrix))
                    if d not in golden:
                        self.fail('Missing @ %s: `Direction.%s: %r,`' % ((amounts, amount), d.name, matrix_tuple))  # noqa: UP031
                    self.assertEqual(golden[d], matrix_tuple, 'Error when calculation coordinates')
        for (ds, amounts, amount) in self.ROTTEN:
            for d in Direction if ds is None else (ds,):
                with self.subTest(d=d, amounts=amounts):
                    with self.assertRaises(ValueError) as e:
                        list(d.multiples(*amounts, amount=amount))
                    logger.info('» %s (%s)%s', d.name, ' '.join((str(a) if a else 'x' for a in amounts)), '[%d]' % amount if amount else '')
                    logger.info('  %s', e.exception)


class Test_VState(unittest.TestCase):
    def test_comparison_invalid(self):
        ii_1 = VState('label', value=None)
        ii_2 = VState('label', value=None)
        ii_3 = VState('other', value=None)
        iv_1 = VState('label', value=1111)
        # Same labels, invalid
        self.assertNotEqual(ii_1, ii_2)
        # Different labels, invalid
        self.assertNotEqual(ii_1, ii_3)
        if __debug__:  # Not a VState
            with self.assertWarns(UserWarning):
                self.assertNotEqual(ii_1, None)
                self.assertNotEqual(ii_1, 1111)
                self.assertNotEqual(iv_1, None)
                self.assertNotEqual(iv_1, 1111)

    def test_comparison_samelabel(self):
        il1_1 = VState('label', value=1)
        il1_2 = VState('label', value=1)
        il2_1 = VState('label', value=2)
        # Same values
        self.assertEqual(il1_1, il1_2)
        # Different values
        self.assertNotEqual(il1_1, il2_1)
        self.assertNotEqual(il1_2, il2_1)

    def test_comparison_difflabel(self):
        i1_1 = VState('one', value=1)
        i2_1 = VState('two', value=1)
        i2_2 = VState('two', value=2)
        # Same values
        self.assertEqual(i1_1, i2_1)
        # Different values
        self.assertNotEqual(i2_1, i2_2)
        self.assertNotEqual(i1_1, i2_2)


class Test_NotebookTabOrder(unittest.TestCase):
    def test_creation_clean(self):
        assert __debug__
        nto_empty = NotebookTabOrder(())
        self.assertEqual(len(nto_empty.shown), 0)
        self.assertEqual(len(nto_empty.disabled), 0)
        nto_basic = NotebookTabOrder(('1', '2', '3'), disabled=set(('2')))
        self.assertEqual(len(nto_basic.shown), 3)
        self.assertEqual(len(nto_basic.disabled), 1)
        self.assertTupleEqual(nto_basic.active, ('1', '3'))
        nto_lists = NotebookTabOrder(['1', '2', '3'], disabled=['3'])
        self.assertIsInstance(nto_lists.shown, typing.Tuple)
        self.assertIsInstance(nto_lists.disabled, typing.Set)
        self.assertTupleEqual(nto_lists.active, ('1', '2'))

    def test_creation_error(self):
        assert __debug__
        # Repeated `shown`
        with self.assertRaises(AssertionError):
            NotebookTabOrder(['1', '2', '1'])
        # Bad disabled keys
        with self.assertRaises(AssertionError):
            NotebookTabOrder(('1', '2', '3'), disabled=set('a'))


if __name__ == '__main__':
    import sys
    logs_lvl = logging.DEBUG if '-v' in sys.argv else logging.INFO
    logging.basicConfig(level=logs_lvl, format='%(levelname)5.5s:%(funcName)s: %(message)s', stream=sys.stderr)
    unittest.main()



# import typing
import unittest
import logging

from tkmilan import util


class Test_Containers(unittest.TestCase):
    def test_ReorderableDict(self):
        d2 = util.ReorderableDict({n: str(n) for n in range(5)})

        # __len__
        self.assertEqual(len(util.ReorderableDict()), 0)
        self.assertEqual(len(d2), 5)
        # __iter__
        self.assertSetEqual(set(k for k in d2), set(range(5)))
        # keys
        self.assertListEqual(list(d2.keys()), list(range(5)))
        # values
        self.assertListEqual(list(d2.values()), list(map(str, range(5))))
        # items
        self.assertListEqual(list(d2.items()), list((n, str(n)) for n in range(5)))

        # index
        self.assertEqual(d2.index(4), 4)
        self.assertEqual(d2.index(10), None)
        # at
        self.assertEqual(d2.at(0), 0)
        self.assertEqual(d2.at(-1), 4)
        for n in range(len(d2)):
            self.assertEqual(d2.index(d2.at(n)), n)
        self.assertListEqual(list(d2.at(slice(0, 2))), [0, 1])
        self.assertListEqual(list(d2.at(slice(-2, None))), [3, 4])

        dc = util.ReorderableDict(d2)
        # __getitem__
        self.assertEqual(dc[4], '4')
        # clear
        dc.clear()
        self.assertEqual(len(dc), 0)
        # get
        self.assertEqual(dc.get(100, None), None)
        # __setitem__
        dc[5] = 'five'
        self.assertEqual(len(dc), 1)
        self.assertEqual(dc[5], 'five')
        self.assertEqual(dc.index(5), 0)
        # __delitem__
        dc[1] = 'one'
        self.assertEqual(dc.index(1), 1)
        del dc[1]
        self.assertEqual(len(dc), 1)

    def test_ReorderableDict_insert(self):
        di = util.ReorderableDict({n: str(n) for n in range(5)})
        di.insert(1, key=10, value='10')
        self.assertEqual(len(di), 6)
        self.assertEqual(di[10], '10')
        self.assertListEqual(list(di.keys()), [0, 10, 1, 2, 3, 4])
        di.insert(4, key=20, value='20')
        self.assertEqual(len(di), 7)
        self.assertEqual(di[20], '20')
        self.assertListEqual(list(di.keys()), [0, 10, 1, 2, 20, 3, 4])

    def test_ReorderableDict_move(self):
        dm = util.ReorderableDict({n: str(n) for n in range(5)})
        dm.move(0, key=4)
        self.assertListEqual(list(dm.items()), list((n, str(n)) for n in [4, 0, 1, 2, 3]))
        dm.move(4, key=0)
        self.assertListEqual(list(dm.items()), list((n, str(n)) for n in [4, 1, 2, 3, 0]))


if __name__ == '__main__':
    import sys
    logs_lvl = logging.DEBUG if '-v' in sys.argv else logging.INFO
    logging.basicConfig(level=logs_lvl, format='%(levelname)5.5s:%(funcName)s: %(message)s', stream=sys.stderr)
    unittest.main()
